@@ -16,10 +16,11 @@ def create_exec_file(id_list, command_template, indir, outdir, output_filename_t
             exec_file.write('o{0}:\t{1}\n'.format(str(ip).zfill(6), ols))
 
 
-def check_remote_path(remote_machine, root_path):
+def check_remote_path(remote_machine, root_path, inverse=False):
     """Checks existence of parent folder of remote root path
     and non-existence of remote root path itself"""
 
+    print("ssh", remote_machine, "ls", root_path)
     lsdir = subprocess.Popen(["ssh", remote_machine, "ls", root_path], stdout=open('/dev/null', 'w'), stderr=subprocess.PIPE).stderr.read().decode('ascii').strip()
     lsdirup = subprocess.Popen(["ssh", remote_machine, "ls", os.path.dirname(root_path[:-1])], stdout=open('/dev/null', 'w'), stderr=subprocess.PIPE).stderr.read().decode('ascii').strip()
     # If parent directory of root path is not there, error
@@ -29,16 +30,21 @@ def check_remote_path(remote_machine, root_path):
         print(lsdirup)
         exit(1)
     # If root path already there, error: you must delete it yourself
-    elif "ls: cannot access" not in lsdir: # If ls on the exec directory does not give error
+    elif "ls: cannot access" not in lsdir and (not inverse): # If ls on the exec directory does not give error
         print(('Exec path {0} is already present in remote location {1}\n'
             'No permission to overwrite it, please delete it manually.') \
             .format(root_path, remote_machine))
         print('ssh {1} rm -rf {0}'.format(root_path, remote_machine))
         exit(1)
+    elif "ls: cannot access" in lsdir and inverse:
+        print(('Exec path {0} is not present in remote location {1}\n'
+            'Rerun cannot start.')\
+            .format(root_path, remote_machine))
+        exit(1)
 
-
-def generate_exec_filesystem(protocol_triad, cache_dir, job_data, runtime_root_path, batch_job_code, 
-        data_transfer_protocol, env_instr=None, build_envroot=None, only_gather=False):
+def generate_exec_filesystem(protocol_triad, cache_dir, job_data, shared_data, runtime_root_path, 
+        batch_job_code, data_transfer_protocol, env_instr=None, build_envroot=None, 
+        only_gather=False):
     
     devnull = open('/dev/null', 'w')
 
@@ -97,32 +103,52 @@ def generate_exec_filesystem(protocol_triad, cache_dir, job_data, runtime_root_p
         os.mkdir(fs_locations["build_exec"])
     if not os.path.exists(fs_locations["build_work"]):
         os.mkdir(fs_locations["build_work"])
-        
-    exec_filesystem = {}  # For each task contains remote work path and shared files paths
+
+    # Must contain: 'batch_paths' -> list ; 'shared_path' 
+    task_folders = {}  # For each task contains remote work path and shared files paths
     shared = {}
     # Step 2: create Shared sub-filesystem -------------------
     for jdi, jd in enumerate(job_data):
-        batchno = jdi // 10000
-        exec_filesystem[jd['unique_code']] = [None, {}]
-
-        for skey in jd['shared_inps']:
+        replacements = {}
+        for skey in jd['shared_inps']:  # dictionary of skey->couples (sbid, path) where sbid is the batch dir index
             if skey not in shared:
                 # Create local batch dir, and copy file there
-                batchisp = (len(shared) + 1)//10000
-                build_shbatch_folder = build_shared_path + 'batch_' + str(batchisp) + '/'  # NOTICE: build_shared_path is local
+                sbid = jd['shared_inps'][skey][0]
+                build_shbatch_folder = build_shared_path + 'batch_' + str(sbid) + '/'  # NOTICE: build_shared_path is local
                 if not os.path.exists(build_shbatch_folder):
                     os.mkdir(build_shbatch_folder)
-                build_shdest_path = build_shbatch_folder + os.path.basename(jd['shared_inps'][skey])
-                shutil.copyfile(jd['shared_inps'][skey], build_shdest_path)
+                build_shdest_path = build_shbatch_folder + os.path.basename(jd['shared_inps'][skey][1])
+                shutil.copyfile(jd['shared_inps'][skey][1], build_shdest_path)
 
                 # Runtime shared destination dir name
                 runtime_shdest_path = (fs_locations["runtime_shared"]
-                        + 'batch_' + str(batchisp) + '/' + os.path.basename(jd['shared_inps'][skey]))
+                        + 'batch_' + str(sbid) + '/' + os.path.basename(jd['shared_inps'][skey][1]))
 
-                shared[skey] = runtime_shdest_path  # shared folder path at execution time
+                shared[skey] = sbid  # shared folder path at execution time
             else:
-                runtime_shdest_path = shared[skey]
-            exec_filesystem[jd['unique_code']][1]["<shared>"+skey] = runtime_shdest_path
+                runtime_shdest_path = (fs_locations["runtime_shared"]
+                        + 'batch_' + str(shared[skey]) + '/' + os.path.basename(jd['shared_inps'][skey][1]))
+            replacements["<shared>"+skey] = runtime_shdest_path
+
+    for skey in shared_data:
+        # TO DO FUNCTION
+        # Duplicate code (see above)
+        if skey not in shared:
+            sbid = shared_data[skey][0]
+            build_shbatch_folder = build_shared_path + 'batch_' + str(sbid) + '/'
+            if not os.path.exists(build_shbatch_folder):
+                os.mkdir(build_shbatch_folder)
+            build_shdest_path = build_shbatch_folder + os.path.basename(shared_data[skey][1])
+            shutil.copyfile(shared_data[skey][1], build_shdest_path)
+
+            runtime_shdest_path = (fs_locations["runtime_shared"]
+                    + 'batch_' + str(sbid) + '/' + os.path.basename(shared_data[skey][1]))
+
+            shared[skey] = sbid
+        else:
+            runtime_shdest_path = (fs_locations["runtime_shared"]
+                    + 'batch_' + str(shared[skey]) + '/' + os.path.basename(shared_data[skey][1]))
+        replacements["<shared>"+skey] = runtime_shdest_path
 
     if protocol == 'remote' and (not env_and_do_not_replicate) and (not only_gather):
         check_remote_path(remote_machine, fs_locations["runtime_root"])
@@ -188,22 +214,20 @@ def generate_exec_filesystem(protocol_triad, cache_dir, job_data, runtime_root_p
         runtime_task_dir = runtime_cache + 'tasks/'
         os.mkdir(build_task_dir)
         for jdi, jd in enumerate(job_data):
-            batchno = jdi // 10000
-
             # Create local task folders and copy clean env files
-            build_batch_folder = build_task_dir + 'batch_' + str(batchno) + '/'
+            build_batch_folder = build_task_dir + 'batch_' + str(jd['batchno']) + '/'
             if not os.path.exists(build_batch_folder):
                 os.mkdir(build_batch_folder)
-            runtime_batch_folder = runtime_task_dir + 'batch_' + str(batchno) + '/'
+            runtime_batch_folder = runtime_task_dir + 'batch_' + str(jd['batchno']) + '/'
             build_task_path = build_batch_folder + 'task_' + jd['unique_code'] + '.sh'
             runtime_task_path = runtime_batch_folder + 'task_' + jd['unique_code'] + '.sh'
-            exec_filesystem[jd['unique_code']][0] = (runtime_batch_folder, os.path.basename(runtime_task_path))
+            task_folders[jd['unique_code']] = (runtime_batch_folder, os.path.basename(runtime_task_path))
 
             # There, create individual task files
             #  Correctly indent the command, writes it in task.sh and gives it exe privileges
             new_command = beautify_bash_oneliner(
                 "cd {0}; ".format(workdir) + jd['command'], 
-                replacements=exec_filesystem[jd['unique_code']][1])
+                replacements=replacements)
 #            print(new_command)
             with open(build_task_path, "w") as tf:
                 tf.write(new_command)
@@ -212,19 +236,14 @@ def generate_exec_filesystem(protocol_triad, cache_dir, job_data, runtime_root_p
 
     else:  # "locusts" mode: optimized environment for parallel and safe execution
     
-        # Batch folders for every 10000 tasks
-        for i in range(len(job_data)//10000 + 1):
-            batch_folder = fs_locations["build_work"] + 'batch_' + str(i) + '/'
-            if not os.path.exists(batch_folder):
-                os.mkdir(batch_folder)
-    
         # Job folders for individual tasks
         for jdi, jd in enumerate(job_data):
-            batchno = jdi // 10000
-    
             # Work sub-environment --------------------------------------------
             # Create local task folders and copy clean env files
-            job_folder = fs_locations["build_work"] + 'batch_' + str(batchno) + '/' + 'task_' + jd['unique_code'] + '/'
+            batch_folder = fs_locations["build_work"] + 'batch_' + str(jd['batchno']) + '/'
+            if not os.path.exists(batch_folder):
+                os.mkdir(batch_folder)
+            job_folder = batch_folder + 'task_' + jd['unique_code'] + '/'
             if not os.path.exists(job_folder):
                 os.mkdir(job_folder)
             for fpath in jd['clean_env_inps']:
@@ -235,7 +254,7 @@ def generate_exec_filesystem(protocol_triad, cache_dir, job_data, runtime_root_p
             # Correctly indent the command, writes it in task.sh and gives it exe privileges
             new_command = beautify_bash_oneliner(
                 jd['command'], 
-                replacements=exec_filesystem[jd['unique_code']][1])
+                replacements=replacements)
             task_filename = job_folder + "task.sh"
             with open(task_filename, "w") as tf:
                 tf.write(new_command)
@@ -243,12 +262,12 @@ def generate_exec_filesystem(protocol_triad, cache_dir, job_data, runtime_root_p
 
             # Runtime paths for task folders
             rem_jf = (fs_locations["runtime_work"] + 'batch_'
-                    + str(batchno) + '/' + 'task_' + jd['unique_code'] + '/')
-            exec_filesystem[jd['unique_code']][0] = (rem_jf, "task.sh")
-
-
-    return {k : exec_filesystem[k][0] for k in exec_filesystem}, fs_locations
+                    + str(jd['batchno']) + '/' + 'task_' + jd['unique_code'] + '/')
+            task_folders[jd['unique_code']] = (rem_jf, "task.sh")
+            if DEBUG:
+                print("SIZE", jdi, get_obj_size(task_folders))
 
+    return task_folders, fs_locations
 
 def create_manager_scripts(protocol_triad, cache_dir, task_folders, partition,
         cpus_per_node, requested_nodes, batch_job_code, fs_locations, 
@@ -326,38 +345,6 @@ def create_manager_scripts(protocol_triad, cache_dir, task_folders, partition,
     else:
         turnon_mempercpu = "#"
 
-
-    """
-    # Reassess the number of requested nodes for avoiding asking
-    #  for too many nodes. For requiring another node, all cpus of all
-    #  nodes must have at least 10 tasks assigned to them
-    #  WARNING: this does not keep into account the length of a task!
-    if protocol != 'local' and min(tasks_per_node_list) < min_stack_per_core*cpus_per_node:
-        grand = min_stack_per_core*cpus_per_node / max(tasks_per_node_list)
-        print("grand", grand)
-        rqn = (
-            int(requested_nodes / grand)
-            + max(tasks_per_node_list) - min(tasks_per_node_list)
-        )
-        print("rqn", rqn)
-        requested_nodes = max(1, rqn)
-        print("requested_nodes", requested_nodes)
-        tasks_per_node_list = distribute_items_in_fixed_length_list(
-            requested_nodes, 
-            len(taskid_list)
-        )
-        print("new tasks_per_node_list", tasks_per_node_list)
-
-    tasks_per_node_list = [x for x in tasks_per_node_list if x > 0]
-    if len(tasks_per_node_list) < cpus_per_node:
-        print("too many processors have been requested")
-        if protocol != 'local':
-            print("node exclusiveness will be turned off")
-            turnon_exclusiveness = "#"
-        cpus_per_node = len(tasks_per_node_list)
-    """
-
-
     print("Protocol used:", protocol)
     print("Total number of tasks:", len(taskid_list))
     print("Actual number of nodes used:", requested_nodes)
@@ -437,10 +424,14 @@ def create_manager_scripts(protocol_triad, cache_dir, task_folders, partition,
         # Job manager identifier and associated tasks
         tasks_per_job.append((jobid, taskid_list[ik:ik+tasks_per_node]))
 
+        if DEBUG:
+            print("SIZE", jobid, get_obj_size(tasks_per_job))
+
     if only_gather:
         return tasks_per_job, ""
 
     manager_cmd, macname = [], {}
+    manager_cmd.append("(")
     manager_cmd.append("echo MANAGERS")
     for job_id, task_list in tasks_per_job:
         macname[job_id] = "{0}/.manager_activity_check_{1}{2}".format(
@@ -448,22 +439,29 @@ def create_manager_scripts(protocol_triad, cache_dir, task_folders, partition,
             batch_job_code,
             str(job_id).zfill(3)
         )
-        manager_cmd.append(" ".join(["date", "-r", macname[job_id], "+%H:%M", "2>&1"]))
+        manager_cmd.append(" ".join(["date", "-r", macname[job_id], "+%H:%M", "2>periodic_check.err.txt"]))
     manager_cmd.append("echo MACNAMES")
     for job_id, task_list in tasks_per_job:
-        manager_cmd.append(" ".join(["cat", macname[job_id], "2>&1"]))
+        manager_cmd.append(" ".join(["cat", macname[job_id], "2>>periodic_check.err.txt"]))
     manager_cmd.append("echo STATUS")
-    manager_cmd.append('squeue --format="%.18i %.9P %.100j %.8u %.2t %.10M %.6D %R"')
+    manager_cmd.append('squeue --format="%.18i %.9P %.100j %.50u %.2t %.10M %.6D %R"')
     manager_cmd.append("echo TIME")
     manager_cmd.append("echo $(date +%H:%M)")
+    manager_cmd.append(") > {0}".format(fs_locations["runtime_exec"] + 'periodic_check.log.txt'))
     with open(fs_locations["build_exec"] + 'periodic_check.sh', 'w') as cmf:
         cmf.write("\n".join(manager_cmd))
     subprocess.call(["chmod", "777", fs_locations["build_exec"] + 'periodic_check.sh'])
     with open(fs_locations["build_exec"] + 'submit.sh', 'w') as sf:
+        sf.write("rm {0}/submit.log\n".format(fs_locations["runtime_exec"]))
         for job_id, task_list in tasks_per_job:
-            sf.write("sbatch {0}/outer_manager_{1}{2}.slurm\n".format(fs_locations["runtime_exec"],
+            sf.write("RES=$(sbatch {0}/outer_manager_{1}{2}.slurm)\necho \"{1}{2} ${{RES##* }}\" >> {0}/submit.log\n".format(fs_locations["runtime_exec"],
                 batch_job_code, str(job_id).zfill(3)))
     subprocess.call(["chmod", "777", fs_locations["build_exec"] + 'submit.sh'])
+    with open(fs_locations["build_exec"] + 'submit_status.sh', 'w') as sf:
+        sf.write("rm {0}/submit_status.log\n".format(fs_locations["runtime_exec"]))
+        #sf.write("for x in `cat {0}/submit.log | awk '{{print $2}}'`\ndo\n\tSTATUS=`/usr/local/bin/dashboard_cli ${{x}} | grep -A1 partition | awk 'NR==2{{print $3}}'` &\n\tsleep 10\n\techo \"${{x}} ${{STATUS}}\" >> {0}/submit_status.log\ndone\n".format(fs_locations["runtime_exec"]))
+        sf.write("for x in `cat {0}/submit.log | awk '{{print $2}}'`\ndo\n\tSTATUS=`jobhist ${{x}} | grep -A1 Partition | awk 'NR==2{{print $3}}'` \n\techo \"${{x}} ${{STATUS}}\" >> {0}/submit_status.log\ndone &>> {0}/submit_status.err\n".format(fs_locations["runtime_exec"]))
+    subprocess.call(["chmod", "777", fs_locations["build_exec"] + 'submit_status.sh'])
 
     if protocol == 'remote':
         cpall_cmd = ["bash", data_transfer_protocol, fs_locations["build_root"], "{0}:{1}".format(remote_machine, fs_locations["runtime_root"])]
@@ -482,7 +480,7 @@ def create_manager_scripts(protocol_triad, cache_dir, task_folders, partition,
 
 
 def remote_job_control(protocol_triad, batch_job_code, fs_locations, tasks_per_job,
-        waiting_time, only_gather=False):
+        waiting_time, data_transfer_protocol, log_folder, only_gather=False):
     # There must be a passwordless connection between the two machines: to
     # achieve it, ssh-agent and then ssh-add.
     # If the two machines share a folder - and thus there is no need to ssh to
@@ -490,16 +488,17 @@ def remote_job_control(protocol_triad, batch_job_code, fs_locations, tasks_per_j
 
     protocol, remote_machine, hpc_shared_folder = protocol_triad
     devnull = open('/dev/null', 'w')
+    NOCACHE_COUNT_THR = 5
 
     if not only_gather:
         if protocol != 'local':
             chk_cmd = [
                 "ssh",
                 remote_machine,
-                "sleep 60;",
                 "bash",
                 fs_locations["runtime_exec"] + 'submit.sh'
             ]
+            time.sleep(40)
             if DEBUG:
                 print("COMMAND", " ".join(chk_cmd))
                 p = subprocess.Popen(chk_cmd)
@@ -508,12 +507,14 @@ def remote_job_control(protocol_triad, batch_job_code, fs_locations, tasks_per_j
                     stderr=devnull,
                     stdout=devnull
                 )
+            p.wait()
+            time.sleep(20)
         else:
             # If no hpc, there is only one node, i.e. one manager
             mname = 'outer_manager_{0}{1}.slurm'.format(batch_job_code, str(0).zfill(3))
             nhp_cmd = ["nohup", fs_locations["runtime_exec"] + mname]
             if DEBUG:
-                print("COMMAND", nhp_cmd)
+                print("COMMAND", " ".join(nhp_cmd))
                 p = subprocess.Popen(nhp_cmd)
             else:
                 p = subprocess.Popen(nhp_cmd,
@@ -533,6 +534,8 @@ def remote_job_control(protocol_triad, batch_job_code, fs_locations, tasks_per_j
     hasrun = {job_id : False for job_id, task_list in tasks_per_job}
     sc, wc = 0, 0
     first_time = True
+    nocache_count = 0
+    no_entry = False
     while not is_over:
         status = {job_id : '' for job_id, task_list in tasks_per_job}
         if not (only_gather and first_time):
@@ -548,59 +551,135 @@ def remote_job_control(protocol_triad, batch_job_code, fs_locations, tasks_per_j
             if not chrono_on[job_id] and time_start[job_id]:
                 time_end[job_id] = time.time()
 
+        """
         if protocol != 'local':
             manager_cmd = ["ssh", remote_machine, "bash", fs_locations["runtime_exec"] + 'periodic_check.sh']
+            smallscp_cmd = ["bash", data_transfer_protocol, "{0}:{1}".format(remote_machine, fs_locations["runtime_exec"] + 'periodic_check.log.txt'), log_folder]
         else:
             manager_cmd = ["bash", fs_locations["build_exec"] + 'periodic_check.sh']
+            smallscp_cmd = ["cp", "{0}".format(fs_locations["runtime_exec"] + 'periodic_check.log.txt'), log_folder]
 
         if DEBUG:
-            print("COMMAND", manager_cmd)
-            periodic_check_txt = subprocess.Popen(
+            print("COMMAND", " ".join(manager_cmd))
+            p = subprocess.Popen(
                 manager_cmd,
                 stdout=subprocess.PIPE
-            ).stdout.read().decode('utf-8')
+            )
         else:
-            periodic_check_txt = subprocess.Popen(
+            p = subprocess.Popen(
                 manager_cmd, 
                 stderr=devnull, 
                 stdout=subprocess.PIPE
-            ).stdout.read().decode('utf-8')
+           )
+        p.wait()
+        if DEBUG:
+            print("COMMAND", " ".join(smallscp_cmd))
+            p = subprocess.Popen(
+                smallscp_cmd,
+                stdout=subprocess.PIPE
+            )
+        else:
+            p = subprocess.Popen(
+                smallscp_cmd, 
+                stderr=devnull, 
+                stdout=subprocess.PIPE
+           )
+        p.wait()
+        """
 
+        if protocol != 'local':
+            jh_cmd = ["ssh", "-t", remote_machine, "bash", fs_locations["runtime_exec"] + 'submit_status.sh']
+            jhcp_cmd = ["bash", data_transfer_protocol, "{0}:{1}".format(remote_machine, fs_locations["runtime_exec"] + 'submit_status.log'), log_folder]
+        else:
+            jh_cmd = ["bash", fs_locations["build_exec"] + 'submit_status.sh']
+            jhcp_cmd = ["cp", "{0}".format(fs_locations["runtime_exec"] + 'submit_status.log'), log_folder]
+
+
+        """
         mng_lines, mac_lines, status_lines, time_line = False, False, False, False
         mng_times, mac_names = [], []
-        for line in periodic_check_txt.split('\n'):
-            if not line.strip():
-                continue
-            if line == "MANAGERS":
-                mng_lines = True
-                continue
-            elif line == "MACNAMES":
-                mng_lines = False
-                mac_lines = True
-                continue
-            elif line == "STATUS":
-                mac_lines = False
-                status_lines = True
-                continue
-            elif line == "TIME":
-                status_lines = False
-                time_line = True
-                continue
-            if mng_lines:
-                mng_times.append(line.strip())
-            elif mac_lines:
-                mac_names.append(line.strip()) 
-            elif status_lines:
-#                print(line)
-                if line.split()[2] in [batch_job_code + str(x[0]).zfill(3) for x in tasks_per_job]:
-#                    print("YE", line)
-                    status[int(line.split()[2][len(batch_job_code):])] = line.split()[4]
-                    hasrun[int(line.split()[2][len(batch_job_code):])] = True
-            elif time_line:
-                real_time = line.strip()
-                real_time = int(real_time.split(":")[0])*60 + int(real_time.split(":")[1])
+        with open(log_folder + '/periodic_check.log.txt') as logf:
+            print("OPENING", log_folder + '/periodic_check.log.txt')
+            for line in logf:
+                if not line.strip():
+                    continue
+                if line == "MANAGERS":
+                    mng_lines = True
+                    continue
+                elif line == "MACNAMES":
+                    mng_lines = False
+                    mac_lines = True
+                    continue
+                elif line == "STATUS":
+                    mac_lines = False
+                    status_lines = True
+                    continue
+                elif line == "TIME":
+                    status_lines = False
+                    time_line = True
+                    continue
+                if mng_lines:
+                    mng_times.append(line.strip())
+                elif mac_lines:
+                    mac_names.append(line.strip()) 
+                elif status_lines:
+#                    print(line)
+                    if line.split()[2] in [batch_job_code + str(x[0]).zfill(3) for x in tasks_per_job]:
+#                        print("YE", line)
+                        status[int(line.split()[2][len(batch_job_code):])] = line.split()[4]
+                        hasrun[int(line.split()[2][len(batch_job_code):])] = True
+                elif time_line:
+                    real_time = line.strip()
+                    real_time = int(real_time.split(":")[0])*60 + int(real_time.split(":")[1])
+            if (not (mng_times and mac_names)) and (not no_entry):
+                print("STILL NO CACHE FILES: CHECKING STATUS")
+        """
 
+        if DEBUG:
+            print("COMMAND", " ".join(jh_cmd))
+            p = subprocess.Popen(
+                jh_cmd,
+                stdout=subprocess.PIPE
+            )
+        else:
+            p = subprocess.Popen(
+                jh_cmd,
+                stderr=devnull,
+                stdout=subprocess.PIPE
+           )
+        p.wait()
+        time.sleep(30)
+        if DEBUG:
+            print("COMMAND", " ".join(jhcp_cmd))
+            p = subprocess.Popen(
+                jhcp_cmd,
+                stdout=subprocess.PIPE
+            )
+        else:
+            p = subprocess.Popen(
+                jhcp_cmd,
+                stderr=devnull,
+                stdout=subprocess.PIPE
+           )
+        p.wait()
+        
+ 
+        all_completed = True
+        with open(log_folder + "/submit_status.log") as ssf:
+            for fline in ssf:
+                if fline.strip():
+#                    print(fline.strip())
+                    fields = fline.split()
+                    if fields[1] != "COMPLETED":
+                        all_completed = False
+                        nocache_count += 1
+        time.sleep(10)
+        if not (all_completed): #and (nocache_count < NOCACHE_COUNT_THR):
+            print("Incomplete jobs are present. Check number", nocache_count)
+            is_over = False
+            continue
 
+        """        
         for it, t in enumerate(tasks_per_job):
             job_id, task_list = t
 
@@ -627,6 +706,7 @@ def remote_job_control(protocol_triad, batch_job_code, fs_locations, tasks_per_j
             else:
                 print(job_id, "End")
                 time_end[job_id] = job_times[job_id]
+        """
 
     log_txt = "\n\n#LOCUSTS JOBS LOG\n"
     time_now = time.time()
@@ -674,7 +754,7 @@ def gather_results(protocol_triad, cache_dir, job_data, batch_job_code,
     # Check if some job did not even start and adds it to the reschedule set 
     reschedule = set()
     for job_id, task_list in tasks_per_job:
-        # Move the mail exec task files to logs/
+        # Move the main exec task files to logs/
         tfname = "taskfile_{0}{1}.*".format(batch_job_code, str(job_id).zfill(3))
         task_filename = fs_locations["build_exec"] + tfname
         mv_cmd = "mv {0} {1}".format(task_filename, log_dir)
@@ -720,8 +800,10 @@ def gather_results(protocol_triad, cache_dir, job_data, batch_job_code,
             if status in ['running', 'pending']:
                 reschedule.add(task_id)
 
+    print("Reschedule these failed jobs:", reschedule)
+
     # For the jobs that have completed, checks the expected outputs
-    if not build_envroot:
+    if (build_envroot is None) or (not build_envroot):
         completed_with_error = set()
         output_paths = {}
         for jd in job_data:
@@ -788,8 +870,7 @@ def gather_results(protocol_triad, cache_dir, job_data, batch_job_code,
     
         # This will be the new job_data, containing all jobs that
         #  have to be rescheduled
-        output_d = { k['unique_code'] : k 
-            for k in job_data if k['unique_code'] in reschedule }
+        new_job_data = [ k for k in job_data if k['unique_code'] in reschedule ]
     
     elif snapshot:
 #        print("TAKE SNAPSHOT", protocol_triad, fs_locations['build_work'])
@@ -821,27 +902,29 @@ def gather_results(protocol_triad, cache_dir, job_data, batch_job_code,
                             p = subprocess.Popen(cpcmd, stdout=devnull, stderr=devnull)
                         p.wait()
 
-        output_d, output_paths, completed_with_error = {}, None, None
+        new_job_data, output_paths, completed_with_error = {}, None, None
 
     # Remove all repositories
     if not DEBUG:
-        if protocol != 'local':
-            sshrm_cmd = ["ssh", remote_machine, "rm", "-rf", fs_locations['runtime_root']]
-            if DEBUG:
-                print("COMMAND", sshrm_cmd)
-                p = subprocess.Popen(sshrm_cmd)
-            else:
+        if not new_job_data:
+            if protocol != 'local':
+                sshrm_cmd = ["ssh", remote_machine, "rm", "-rf", fs_locations['runtime_root']]
                 p = subprocess.Popen(sshrm_cmd, stderr=devnull, stdout=devnull)
-            p.wait()
-        rm_cmd = ["rm", "-rf", fs_locations['build_root']]
-        if DEBUG:
-            print("COMMAND", rm_cmd)
-            p = subprocess.Popen(rm_cmd)
-        else:
+                p.wait()
+            rm_cmd = ["rm", "-rf", fs_locations['build_root']]
             p = subprocess.Popen(rm_cmd, stderr=devnull, stdout=devnull)
-        p.wait()
+            p.wait()
+        else:
+            print("WARNING - only remove local and remote exec dirs", fs_locations['build_exec'], fs_locations['runtime_exec'])
+            if protocol != 'local':
+                sshrm_cmd = ["ssh", remote_machine, "rm", "-rf", fs_locations['runtime_exec']]
+                p = subprocess.Popen(sshrm_cmd, stderr=devnull, stdout=devnull)
+                p.wait()
+            rm_cmd = ["rm", "-rf", fs_locations['build_exec']]
+            p = subprocess.Popen(rm_cmd, stderr=devnull, stdout=devnull)
+            p.wait()
 
-    return output_d, output_paths, completed_with_error
+    return new_job_data, output_paths, completed_with_error
 
 def take_snapshot(protocol_triad, root_dir):
     protocol, remote_machine, hpc_shared_folder = protocol_triad
@@ -922,8 +1005,12 @@ def highly_parallel_job_manager(options, exec_filename,
     singinfo = (options['singularity'], options['singularity_container'], options['singularity_modload'])
 
     # Read exec file and compile job_data
+    batch_size = options['batch_size']
     job_data = []
     jd = {}
+    jdi = -1
+    sharedwall = {}
+    totsharecount = []
     with open(exec_filename) as exec_file:
         for line in exec_file:
             if line.startswith("c"):
@@ -941,10 +1028,13 @@ def highly_parallel_job_manager(options, exec_filename,
                     'success' : None,
                     'issues' : [],
                     'unique_code' : '',
+                    'batchno' : '',
                     'log_filename' : '',
                     'clean_env_inps' : [],
-                    'shared_inps' : []
+                    'shared_inps' : {}
                 }
+                jdi += 1
+                jd['batchno'] = jdi // batch_size
                 jd['command'] = line[8:].strip()
                 fields = line.split()
                 jd['output_dir'] = locout_dir
@@ -956,11 +1046,31 @@ def highly_parallel_job_manager(options, exec_filename,
                 jd['clean_env_inps'] = fields[1:]
             elif line.startswith("s"):  # shared inputs must be declared in "s" line as file.txt:/path/of/file.txt, where file.txt is a filename appearing in the "c" line
                 fields = line.split()
-                jd['shared_inps'] = {k : p 
-                    for (k,p) in [x.split(":") for x in fields[1:]]}
+                for x in fields[1:]:
+                    k, pth = x.split(":")
+                    if pth in totsharecount:
+                        sbid = totsharecount.index(pth) // batch_size
+                    else:
+                        sbid = len(totsharecount) // batch_size   # should be len(x) +1 (for append) -1 (for index starting from 0)
+                        totsharecount.append(pth)
+                    jd['shared_inps'][k] = (sbid, pth)
             elif line.startswith("o"):
                 fields = line.split()
                 jd['outputs'] = fields[1:]
+            elif line.startswith("S"): # shared inputs that are common to all tasks. Only one S record is allowed in a file
+                fields = line.split()
+                if sharedwall:
+                    print("ERROR: more than one S record in file {0}"\
+                        .format(exec_filename))
+                    exit(1)
+                for x in fields[1:]:
+                    k, pth = x.split(":")
+                    if pth in totsharecount:
+                        sbid = (totsharecount.index(pth) + 1) // batch_size
+                    else:
+                        sbid = (len(totsharecount) + 1) // batch_size
+                        totsharecount.append(pth)
+                    sharedwall[k] = (sbid, pth)
         if jd:
             if os.path.exists(jd['log_filename']):
                 if options['force_redo']:
@@ -1005,6 +1115,8 @@ def highly_parallel_job_manager(options, exec_filename,
     else:
         env_root_dir, gen_env_root_dir = None, None
 
+    #print("ENV ROOT DIR", env_root_dir)
+
     data_transfer_protocol = options['data_transfer_protocol']
     email = options['email_address']
     nsf = options['nodewise_scratch_folder']
@@ -1020,31 +1132,35 @@ def highly_parallel_job_manager(options, exec_filename,
     memory_per_cpu = options['memory_per_cpu']
     only_gather = options['only_gather']
 
-    log_txt = ("#RUN LOG\n\n"
-        "Run type (local / shared / remote) :\t{0}\n").format(protocol)
-    if protocol != 'local':
-        log_txt += ("Remote machine :\t{0}\n"
-            "Partition (queue) :\t{1}\n"
-            "Constraint (node type) :\t{2}\n"
-            "Number of reserved nodes :\t{3}\n"
-            "(Minimum) number of cores per node :\t{4}\n"
-            "Exclusive use of nodes :\t{5}\n"
-            "Tasks :\t{6}\n"
-            "Minimum number of tasks assigned per core :\t{7}\n"
-            "Wall time :\t{8}\n"
-            "Nodewise scratch folder :\t{9}\n"
-            "Nodewise scratch memory :\t{10}\n"
-        ).format(remote_machine, partition, constraint, requested_nodes, cpus_per_node, exclusive, len(job_data), task_stack, wt, nsf, nsm)   
-    else:
-        log_txt += "Number of local processors :\t{0}\n".format(cpus_per_node)
-    print(log_txt)
-
+    rerun = False
+    already_rescheduled = set()
     while job_data:
+        # Initial log
+        log_txt = ("#RUN LOG\n\n"
+            "Run type (local / shared / remote) :\t{0}\n").format(protocol)
+        if protocol != 'local':
+            log_txt += ("Remote machine :\t{0}\n"
+                "Partition (queue) :\t{1}\n"
+                "Constraint (node type) :\t{2}\n"
+                "Number of reserved nodes :\t{3}\n"
+                "(Minimum) number of cores per node :\t{4}\n"
+                "Exclusive use of nodes :\t{5}\n"
+                "Tasks :\t{6}\n"
+                "Minimum number of tasks assigned per core :\t{7}\n"
+                "Wall time :\t{8}\n"
+                "Nodewise scratch folder :\t{9}\n"
+                "Nodewise scratch memory :\t{10}\n"
+            ).format(remote_machine, partition, constraint, requested_nodes, cpus_per_node, exclusive, len(job_data), task_stack, wt, nsf, nsm)   
+        else:
+            log_txt += "Number of local processors :\t{0}\n".format(cpus_per_node)
+        print(log_txt)
+
         # Create the hosting file system
         task_folders, fs_locations = generate_exec_filesystem(
             protocol_triad,
             cache_dir, 
-            job_data, 
+            job_data,
+            sharedwall, 
             runtime_root_path,
             batch_job_code,
             data_transfer_protocol,
@@ -1052,6 +1168,7 @@ def highly_parallel_job_manager(options, exec_filename,
             build_envroot=gen_env_root_dir,
             only_gather=only_gather
         )
+        gc.collect()
 
         # Create local manager script that does the mpiq job, launchable on each node. It checks the situation regularly each 10 secs.
         tasks_per_job, tmp_log_txt = create_manager_scripts(
@@ -1078,6 +1195,7 @@ def highly_parallel_job_manager(options, exec_filename,
             only_gather=only_gather
         )
         log_txt += tmp_log_txt
+        gc.collect()
 
         if env_instr:
             snapshot = take_snapshot(
@@ -1108,12 +1226,15 @@ def highly_parallel_job_manager(options, exec_filename,
             fs_locations, 
             tasks_per_job, 
             waiting_time,
+            data_transfer_protocol,
+            locout_dir,
             only_gather=only_gather
         )
+        gc.collect()
 
 
         # Collect results, update job_data (only processes that remained pending on running are written in job_data again
-        job_data, outp, witherr = gather_results(
+        resc_job_data, outp, witherr = gather_results(
             protocol_triad,
             cache_dir, 
             job_data, 
@@ -1132,6 +1253,20 @@ def highly_parallel_job_manager(options, exec_filename,
         if outp:
             for x in outp:
                 output_paths[x] = outp[x]
+        gc.collect()
+
+        this_rescheduling = []
+        if resc_job_data:
+            for k in resc_job_data:
+                if k['unique_code'] not in already_rescheduled:
+                    this_rescheduling.append(k)
+                    already_rescheduled.add(k['unique_code'])
+
+        if this_rescheduling:
+            job_data = this_rescheduling
+            rerun = True
+        else:
+            job_data = []
 
     run_log_fn = cache_dir + 'run.log'
     with open(run_log_fn, 'w') as rf:
